#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 17 May 2023
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

"""
Code taken from GREMLIN repository available at https://github.com/sokrypton/GREMLIN_CPP/
and adapted (put functions into a class termed GREMLIN) and used under the
"THE BEER-WARE LICENSE" (Revision 42):
 ----------------------------------------------------------------------------------------
 "THE BEER-WARE LICENSE" (Revision 42):
 <so@fas.harvard.edu> wrote this file.  As long as you retain this notice you
 can do whatever you want with this stuff. If we meet some day, and you think
 this stuff is worth it, you can buy me a beer in return.  Sergey Ovchinnikov
 ----------------------------------------------------------------------------------------
--> Thanks for sharing the great code, I will gladly provide you a beer or two. (Niklas)
Code mainly taken from
https://github.com/sokrypton/GREMLIN_CPP/blob/master/GREMLIN_TF.ipynb

References:
[1] Kamisetty, H., Ovchinnikov, S., & Baker, D.
    Assessing the utility of coevolution-based residue–residue contact predictions in a
    sequence- and structure-rich era.
    Proceedings of the National Academy of Sciences, 2013, 110, 15674-15679
    https://www.pnas.org/doi/10.1073/pnas.1314045110
[2] Balakrishnan, S., Kamisetty, H., Carbonell, J. G., Lee, S.-I., & Langmead, C. J.
    Learning generative models for protein fold families.
    Proteins, 79(4), 2011, 1061–78.
    https://doi.org/10.1002/prot.22934
[3] Ekeberg, M., Lövkvist, C., Lan, Y., Weigt, M., & Aurell, E.
    Improved contact prediction in proteins: Using pseudolikelihoods to infer Potts models.
    Physical Review E, 87(1), 2013, 012707. doi:10.1103/PhysRevE.87.012707
    https://doi.org/10.1103/PhysRevE.87.012707
"""

import logging
logger = logging.getLogger('pypef.dca.params_inference')

from os import mkdir
import pickle
import numpy as np
import matplotlib.pyplot as plt
from scipy.spatial.distance import pdist, squareform
from scipy.special import logsumexp
from scipy.stats import boxcox
import pandas as pd
import tensorflow as tf
tf.get_logger().setLevel('DEBUG')

from pypef.utils.variant_data import get_sequences_from_file


class GREMLIN:
    def __init__(
            self,
            alignment: str,
            char_alphabet: str = "ARNDCQEGHILKMFPSTWYV-",
            wt_seq=None,
            optimize=True,
            gap_cutoff=0.5,
            eff_cutoff=0.8,
            opt_iter=100
    ):
        self.char_alphabet = char_alphabet
        self.gap_cutoff = gap_cutoff
        self.eff_cutoff = eff_cutoff
        self.opt_iter = opt_iter
        self.states = len(self.char_alphabet)
        self.a2n = self.a2n_dict()
        self.seqs, _, _ = get_sequences_from_file(alignment)
        self.msa_ori = self.get_msa_ori()
        self.n_col_ori = self.msa_ori.shape[1]
        if wt_seq is not None:
            self.wt_seq = wt_seq
        else:  # Taking the first sequence in the MSA as wild type sequence
            logger.info("No wild-type sequence provided: The first sequence "
                        "in the MSA is considered the wild-type sequence.")
            self.wt_seq = "".join([self.char_alphabet[i] for i in self.msa_ori[0]])
        if len(self.wt_seq) != self.n_col_ori:
            raise SystemError("Length of (provided) wild-type sequence does not match "
                              "number of MSA columns, i.e., common MSA sequence length.")
        self.msa_trimmed, self.v_idx, self.w_idx, self.w_rel_idx, self.gaps = self.filt_gaps(self.msa_ori)
        self.msa_weights = self.get_eff_msa_weights(self.msa_trimmed)
        self.n_eff = np.sum(self.msa_weights)
        self.n_row = self.msa_trimmed.shape[0]
        self.n_col = self.msa_trimmed.shape[1]
        self.v_ini, self.w_ini = self.initialize_v_w(remove_gap_entries=False)
        self.optimize = optimize
        if self.optimize:
            self.v_opt, self.w_opt = self.run_opt_tf()
        self.x_wt = self.collect_encoded_sequences(np.atleast_1d(self.wt_seq))

    def a2n_dict(self):
        a2n = {}
        for a, n in zip(self.char_alphabet, range(self.states)):
            a2n[a] = n
        return a2n

    def aa2int(self, aa):
        """convert single aa into numerical integer value, e.g.
        "A" -> 0 or "-" to 21 dependent on char_alphabet"""
        if aa in self.a2n:
            return self.a2n[aa]
        else:  # for unknown characters insert Gap character
            return self.a2n['-']

    def str2int(self, x):
        """
        convert a list of strings into list of integers
        Example: ["ACD","EFG"] -> [[0,4,3], [6,13,7]]
        """
        if type(x) == list:
            x = np.array(x)
        if x.dtype.type is np.str_:
            if x.ndim == 0:  # single seq
                return np.array([self.aa2int(aa) for aa in str(x)])
            else:  # list of seqs
                return np.array([[self.aa2int(aa) for aa in seq] for seq in x])
        else:
            return x

    @property
    def get_v_idx_w_idx(self):
        return self.v_idx, self.w_idx

    def get_msa_ori(self):
        """converts list of sequences to msa"""
        msa_ori = []
        for seq in self.seqs:
            msa_ori.append([self.aa2int(aa.upper()) for aa in seq])
        msa_ori = np.array(msa_ori)
        return msa_ori

    def filt_gaps(self, msa_ori):
        """filters alignment to remove gappy positions"""
        tmp = (msa_ori == self.states - 1).astype(float)
        non_gaps = np.where(np.sum(tmp.T, -1).T / msa_ori.shape[0] < self.gap_cutoff)[0]

        gaps = np.where(np.sum(tmp.T, -1).T / msa_ori.shape[0] >= self.gap_cutoff)[0]
        logger.info(f'Gap positions (removed from msa):\n{gaps}')
        ncol_trimmed = len(non_gaps)
        v_idx = non_gaps
        w_idx = v_idx[np.stack(np.triu_indices(ncol_trimmed, 1), -1)]
        w_rel_idx = np.stack(np.triu_indices(ncol_trimmed, 1), -1)
        return msa_ori[:, non_gaps], v_idx, w_idx, w_rel_idx, gaps

    def get_eff_msa_weights(self, msa):
        """compute effective weight for each sequence"""
        # pairwise identity
        pdistance_msa = pdist(msa, "hamming")
        msa_sm = 1.0 - squareform(pdistance_msa)
        # weight for each sequence
        msa_w = (msa_sm >= self.eff_cutoff).astype(float)
        msa_w = 1 / np.sum(msa_w, -1)
        return msa_w

    @staticmethod
    def l2(x):
        return np.sum(np.square(x))

    def objective(self, v, w=None, flattened=True):
        """Same objective function as used in run_opt_tf below
        but here only using numpy not TensorFlow functions.
        Potentially helpful for implementing SciPy optimizers."""
        if w is None:
            w = self.w_ini
        onehot_cat_msa = np.eye(self.states)[self.msa_trimmed]
        if flattened:
            v = np.reshape(v, (self.n_col, self.states))
            w = np.reshape(w, (self.n_col, self.states, self.n_col, self.states))
        ########################################
        # Pseudo-Log-Likelihood
        ########################################
        # v + w
        vw = v + np.tensordot(onehot_cat_msa, w, 2)
        # Hamiltonian
        h = np.sum(np.multiply(onehot_cat_msa, vw), axis=(1, 2))
        # local z (partition function)
        z = np.sum(np.log(np.sum(np.exp(vw), axis=2)), axis=1)
        # Pseudo-Log-Likelihood
        pll = h - z
        ########################################
        # Regularization
        ########################################
        l2_v = 0.01 * self.l2(v)
        l2_w = 0.01 * self.l2(w) * 0.5 * (self.n_col - 1) * (self.states - 1)
        # loss function to minimize
        loss = -np.sum(pll * self.msa_weights) / np.sum(self.msa_weights)
        loss = loss + (l2_v + l2_w) / self.n_eff
        return loss

    @staticmethod
    def opt_adam(loss, name, var_list=None, lr=1.0, b1=0.9, b2=0.999, b_fix=False):
        """
        Adam optimizer [https://arxiv.org/abs/1412.6980] with first and second moments
            mt          and
            vt          (greek letter nu) at time steps t, respectively.
        Note by GREMLIN authors: this is a modified version of adam optimizer.
        More specifically, we replace "vt" with sum(g*g) instead of (g*g).
        Furthermore, we find that disabling the bias correction
        (b_fix=False) speeds up convergence for our case.
        """
        if var_list is None:
            var_list = tf.compat.v1.trainable_variables()
        gradients = tf.gradients(loss, var_list)
        if b_fix:
            t = tf.Variable(0.0, "t")
        opt = []
        for n, (x, g) in enumerate(zip(var_list, gradients)):
            if g is not None:
                ini = dict(initializer=tf.zeros_initializer, trainable=False)
                mt = tf.compat.v1.get_variable(name + "_mt_" + str(n), shape=list(x.shape), **ini)
                vt = tf.compat.v1.get_variable(name + "_vt_" + str(n), shape=[], **ini)

                mt_tmp = b1 * mt + (1 - b1) * g
                vt_tmp = b2 * vt + (1 - b2) * tf.reduce_sum(tf.square(g))
                lr_tmp = lr / (tf.sqrt(vt_tmp) + 1e-8)

                if b_fix:
                    lr_tmp = lr_tmp * tf.sqrt(1 - tf.pow(b2, t)) / (1 - tf.pow(b1, t))

                opt.append(x.assign_add(-lr_tmp * mt_tmp))
                opt.append(vt.assign(vt_tmp))
                opt.append(mt.assign(mt_tmp))

        if b_fix:
            opt.append(t.assign_add(1.0))
        return tf.group(opt)

    @staticmethod
    def sym_w(w):
        """
        Symmetrize input matrix of shape (x,y,x,y)
        As the full couplings matrix W might/will be slightly "unsymmetrical"
        it will be symmetrized according to one half being "mirrored".
        """
        x = w.shape[0]
        w = w * np.reshape(1 - np.eye(x), (x, 1, x, 1))
        w = w + tf.transpose(w, [2, 3, 0, 1])
        return w

    @staticmethod
    def l2_tf(x):
        return tf.reduce_sum(tf.square(x))

    def run_opt_tf(self, opt_rate=1.0, batch_size=None):
        """
        For optimization of v and w ADAM is used here (L-BFGS-B not (yet) implemented
        for TF 2.x, e.g. using scipy.optimize.minimize).
        Gaps (char '-' respectively '21') included.
        """
        ##############################################################
        # SETUP COMPUTE GRAPH
        ##############################################################
        # kill any existing tensorflow graph
        tf.compat.v1.reset_default_graph()
        tf.compat.v1.disable_eager_execution()

        # msa (multiple sequence alignment)
        msa = tf.compat.v1.placeholder(tf.int32, shape=(None, self.n_col), name="msa")

        # one-hot encode msa
        oh_msa = tf.one_hot(msa, self.states)

        # msa weights
        msa_weights = tf.compat.v1.placeholder(tf.float32, shape=(None,), name="msa_weights")

        # 1-body-term of the MRF
        v = tf.compat.v1.get_variable(name="v",
                                      shape=[self.n_col, self.states],
                                      initializer=tf.compat.v1.zeros_initializer)

        # 2-body-term of the MRF
        w = tf.compat.v1.get_variable(name="w",
                                      shape=[self.n_col, self.states, self.n_col, self.states],
                                      initializer=tf.compat.v1.zeros_initializer)

        # symmetrize w
        w = self.sym_w(w)

        ########################################
        # Pseudo-Log-Likelihood
        ########################################
        # v + w
        vw = v + tf.tensordot(oh_msa, w, 2)

        # Hamiltonian
        h = tf.reduce_sum(tf.multiply(oh_msa, vw), axis=(1, 2))
        # partition function Z
        z = tf.reduce_sum(tf.reduce_logsumexp(vw, axis=2), axis=1)

        # Pseudo-Log-Likelihood
        pll = h - z

        ########################################
        # Regularization
        ########################################
        l2_v = 0.01 * self.l2_tf(v)
        lw_w = 0.01 * self.l2_tf(w) * 0.5 * (self.n_col - 1) * (self.states - 1)

        # loss function to minimize
        loss = -tf.reduce_sum(pll * msa_weights) / tf.reduce_sum(msa_weights)
        loss = loss + (l2_v + lw_w) / self.n_eff

        ##############################################################
        # MINIMIZE LOSS FUNCTION
        ##############################################################
        opt = self.opt_adam(loss, "adam", lr=opt_rate)
        # initialize V (local fields)
        msa_cat = tf.keras.utils.to_categorical(self.msa_trimmed, self.states)
        pseudo_count = 0.01 * np.log(self.n_eff)
        v_ini = np.log(np.sum(msa_cat.T * self.msa_weights, -1).T + pseudo_count)
        v_ini = v_ini - np.mean(v_ini, -1, keepdims=True)

        # generate input/feed
        def feed(feed_all=False):
            if batch_size is None or feed_all:
                return {msa: self.msa_trimmed, msa_weights: self.msa_weights}
            else:
                idx = np.random.randint(0, self.n_row, size=batch_size)
                return {msa: self.msa_trimmed[idx], msa_weights: self.msa_weights[idx]}

        with tf.compat.v1.Session() as sess:
            # initialize variables V (local fields) and W (couplings)
            sess.run(tf.compat.v1.global_variables_initializer())
            sess.run(v.assign(v_ini))
            # compute loss across all data
            get_loss = lambda: round(sess.run(loss, feed(feed_all=True)) * self.n_eff, 2)
            logger.info(f"Initial loss: {get_loss()}. Starting parameter optimization...")
            for i in range(self.opt_iter):
                sess.run(opt, feed())
                try:
                    if (i + 1) % int(self.opt_iter / 10) == 0:
                        logger.info(f"Iteration {(i + 1)} {get_loss()}")
                except ZeroDivisionError:
                    logger.info(f"Iteration {(i + 1)} {get_loss()}")
            # save the v and w parameters of the MRF
            v_opt = sess.run(v)
            w_opt = sess.run(w)

        no_gap_states = self.states - 1
        return v_opt[:, :no_gap_states], w_opt[:, :no_gap_states, :, :no_gap_states]

    def initialize_v_w(self, remove_gap_entries=True):
        """
        For optimization of v and w ADAM is used here (L-BFGS-B not (yet)
        implemented for TF 2.x, e.g. using scipy.optimize.minimize).
        Gaps (char '-' respectively '21') included.
        """
        w_ini = np.zeros((self.n_col, self.states, self.n_col, self.states))
        onehot_cat_msa = np.eye(self.states)[self.msa_trimmed]
        pseudo_count = 0.01 * np.log(self.n_eff)
        v_ini = np.log(np.sum(onehot_cat_msa.T * self.msa_weights, -1).T + pseudo_count)
        v_ini = v_ini - np.mean(v_ini, -1, keepdims=True)
        # loss_score_ini = self.objective(v_ini, w_ini, flattened=False)  # * self.n_eff

        if remove_gap_entries:
            no_gap_states = self.states - 1
            v_ini = v_ini[:, :no_gap_states]
            w_ini = w_ini[:, :no_gap_states, :, :no_gap_states]

        return v_ini, w_ini

    @property
    def get_v_w_opt(self):
        try:
            return self.v_opt, self.w_opt
        except AttributeError:
            raise SystemError(
                "No v_opt and w_opt available, this means GREMLIN "
                "has not been initialized setting optimize to True, "
                "e.g., try GREMLIN('Alignment.fasta', optimize=True)."
            )

    def get_score(self, seqs, v=None, w=None, v_idx=None, encode=False, h_wt_seq=0.0, recompute_z=False):
        if v is None or w is None:
            if self.optimize:
                v, w = self.v_opt, self.w_opt
            else:
                v, w = self.v_ini, self.w_ini
        if v_idx is None:
            v_idx = self.v_idx
        seqs_int = self.str2int(seqs)
        # if length of sequence != length of model use only
        # valid positions (v_idx) from the trimmed alignment
        try:
            if seqs_int.shape[-1] != len(v_idx):
                seqs_int = seqs_int[..., v_idx]
        except IndexError:
            raise SystemError(
                "The loaded GREMLIN parameter model does not match the input model "
                "in terms of sequence encoding shape or is a gap-substituted sequence. "
                "E.g., when providing two different DCA models/parameters provided by: "
                "\"-m DCA\" and \"--params GREMLIN\", where -m DCA represents a ml input "
                "model potentially generated using plmc parameters and --params GREMLIN "
                "provides differently encoded sequences generated using GREMLIN."
            )

        # one hot encode
        x = np.eye(self.states)[seqs_int]
        # aa_pos_counts = np.sum(x, axis=0)

        # get non-gap positions
        # no_gap = 1.0 - x[..., -1]

        # remove gap from one-hot-encoding
        x = x[..., :-1]

        # compute score
        vw = v + np.tensordot(x, w, 2)

        # ============================================================================================
        # Note, Z (the partition function) is a constant. In GREMLIN, V, W & Z are estimated using all
        # the original weighted input sequence(s). It is NOT recommended to recalculate z with a
        # different set of sequences. Given the common ERROR of recomputing Z, we include the option
        # to do so, for comparison.
        # ============================================================================================
        h = np.sum(np.multiply(x, vw), axis=-1)

        if encode:
            return h

        if recompute_z:
            z = logsumexp(vw, axis=-1)
            return np.sum((h - z), axis=-1) - h_wt_seq
        else:
            return np.sum(h, axis=-1) - h_wt_seq

    def get_wt_score(self, wt_seq=None, v=None, w=None):
        if wt_seq is None:
            wt_seq = self.wt_seq
        if v is None or w is None:
            if self.optimize:
                v, w = self.v_opt, self.w_opt
            else:
                v, w = self.v_ini, self.w_ini
        wt_seq = np.array(wt_seq, dtype=str)
        return self.get_score(wt_seq, v, w)

    def collect_encoded_sequences(self, seqs, v=None, w=None, v_idx=None):
        """
        Wrapper function for encoding input sequences using the self.get_score
        function with encode set to True.
        """
        xs = self.get_score(seqs, v, w, v_idx, encode=True)
        return xs

    @staticmethod
    def normalize(apc_mat):
        """
        Normalization of APC matrix for getting z-Score matrix
        """
        dim = apc_mat.shape[0]
        apc_mat_flat = apc_mat.flatten()
        x, _ = boxcox(apc_mat_flat - np.amin(apc_mat_flat) + 1.0)
        x_mean = np.mean(x)
        x_std = np.std(x)
        x = (x - x_mean) / x_std
        x = x.reshape(dim, dim)
        return x

    def mtx_gaps_as_zeros(self, gap_reduced_mtx, insert_gap_zeros=True):
        """
        Inserts zeros at gap positions of the (L,L) matrices,
        i.e., raw/apc/zscore matrices.
        """
        if insert_gap_zeros:
            gap_reduced_mtx = list(gap_reduced_mtx)
            mtx_zeroed = []
            c = 0
            for i in range(self.n_col_ori):
                mtx_i = []
                c_i = 0
                if i in self.gaps:
                    mtx_zeroed.append(list(np.zeros(self.n_col_ori)))
                else:
                    for j in range(self.n_col_ori):
                        if j in self.gaps:
                            mtx_i.append(0.0)
                        else:
                            mtx_i.append(gap_reduced_mtx[c][c_i])
                            c_i += 1
                    mtx_zeroed.append(mtx_i)
                    c += 1
            return np.array(mtx_zeroed)

        else:
            return gap_reduced_mtx

    def get_correlation_matrix(self, matrix_type: str = 'apc', insert_gap_zeros=False):
        """
        Requires optimized w matrix (of shape (L, 20, L, 20))
        inputs
        ------------------------------------------------------
        w           : coevolution       shape=(L,A,L,A)
        ------------------------------------------------------
        outputs
        ------------------------------------------------------
        raw         : l2norm(w)         shape=(L,L)
        apc         : apc(raw)          shape=(L,L)
        zscore      : normalize(apc)    shape=(L,L)
        """
        # l2norm of 20x20 matrices (note: gaps already excluded)
        raw = np.sqrt(np.sum(np.square(self.w_opt), (1, 3)))

        # apc (average product correction)
        ap = np.sum(raw, 0, keepdims=True) * np.sum(raw, 1, keepdims=True) / np.sum(raw)
        apc = raw - ap

        if matrix_type == 'apc':
            return self.mtx_gaps_as_zeros(apc, insert_gap_zeros=insert_gap_zeros)
        elif matrix_type == 'raw':
            return self.mtx_gaps_as_zeros(raw, insert_gap_zeros=insert_gap_zeros)
        elif matrix_type == 'zscore' or matrix_type == 'z_score':
            return self.mtx_gaps_as_zeros(self.normalize(apc), insert_gap_zeros=insert_gap_zeros)
        else:
            raise SystemError("Unknown matrix type. Choose between 'apc', 'raw', or 'zscore'.")

    def plot_correlation_matrix(self, matrix_type: str = 'apc', set_diag_zero=True):
        matrix = self.get_correlation_matrix(matrix_type, insert_gap_zeros=True)
        if set_diag_zero:
            np.fill_diagonal(matrix, 0.0)

        fig, ax = plt.subplots(figsize=(10, 10))

        if matrix_type == 'zscore' or matrix_type == 'z_score':
            ax.imshow(matrix, cmap='Blues', interpolation='none', vmin=1, vmax=3)
        else:
            ax.imshow(matrix, cmap='Blues')
        tick_pos = ax.get_xticks()
        tick_pos[-1] = matrix.shape[0]
        tick_pos[2:] -= 1
        ax.set_xticks(tick_pos)
        ax.set_yticks(tick_pos)
        labels = [item.get_text() for item in ax.get_xticklabels()]
        labels = [labels[0]] + [str(int(label) + 1) for label in labels[1:]]
        ax.set_xticklabels(labels)
        ax.set_yticklabels(labels)
        ax.set_xlim(-1, matrix.shape[0])
        ax.set_ylim(-1, matrix.shape[0])
        plt.title(matrix_type)
        plt.savefig(f'{matrix_type}.png', dpi=500)
        plt.close('all')

    def get_top_coevolving_residues(self, wt_seq=None, min_distance=0, sort_by="apc"):
        if wt_seq is None:
            wt_seq = self.wt_seq
        if wt_seq is None:
            raise SystemError("Getting top co-evolving residues requires "
                              "the wild type sequence as input.")
        raw = self.get_correlation_matrix(matrix_type='raw')
        apc = self.get_correlation_matrix(matrix_type='apc')
        zscore = self.get_correlation_matrix(matrix_type='zscore')

        # Explore top co-evolving residue pairs
        i_rel_idx = self.w_rel_idx[:, 0]
        j_rel_idx = self.w_rel_idx[:, 1]

        apc_flat = []
        zscore_flat = []
        raw_flat = []
        for i, _ in enumerate(i_rel_idx):
            raw_flat.append(raw[i_rel_idx[i]][j_rel_idx[i]])
            apc_flat.append(apc[i_rel_idx[i]][j_rel_idx[i]])
            zscore_flat.append(zscore[i_rel_idx[i]][j_rel_idx[i]])

        i_idx = self.w_idx[:, 0]
        j_idx = self.w_idx[:, 1]

        i_aa = [f"{wt_seq[i]}_{i + 1}" for i in i_idx]
        j_aa = [f"{wt_seq[j]}_{j + 1}" for j in j_idx]

        # load mtx into pandas dataframe
        mtx = {
            "i": i_idx, "j": j_idx, "apc": apc_flat, "zscore": zscore_flat,
            "raw": raw_flat, "i_aa": i_aa, "j_aa": j_aa
        }
        df_mtx = pd.DataFrame(mtx, columns=["i", "j", "apc", "zscore", "raw", "i_aa", "j_aa"])
        df_mtx_sorted = df_mtx.sort_values(sort_by, ascending=False)

        # get contacts with sequence separation > min_distance
        df_mtx_sorted_mindist = df_mtx_sorted.loc[df_mtx['j'] - df_mtx['i'] > min_distance]

        return df_mtx_sorted_mindist


"""
GREMLIN class helper functions below.
"""


def save_gremlin_as_pickle(alignment: str, wt_seq: str, opt_iter: int = 100):
    """
    Function for getting and/or saving (optimized or unoptimized) GREMLIN model
    """
    logger.info(f'Inferring GREMLIN DCA parameters based on the provided MSA...')
    gremlin = GREMLIN(alignment, wt_seq=wt_seq, optimize=True, opt_iter=opt_iter)
    try:
        mkdir('Pickles')
    except FileExistsError:
        pass

    logger.info(f'Saving GREMLIN model as Pickle file...')
    pickle.dump(
        {
            'model': gremlin,
            'model_type': 'GREMLINpureDCA',
            'beta_1': None,
            'beta_2': None,
            'spearman_rho': None,
            'regressor': None
        },
        open('Pickles/GREMLIN', 'wb')
    )
    return gremlin


def plot_all_corr_mtx(gremlin: GREMLIN):
    gremlin.plot_correlation_matrix(matrix_type='raw')
    gremlin.plot_correlation_matrix(matrix_type='apc')
    gremlin.plot_correlation_matrix(matrix_type='zscore')


def save_corr_csv(gremlin: GREMLIN, min_distance: int = 0, sort_by: str = 'apc'):
    df_mtx_sorted_mindist = gremlin.get_top_coevolving_residues(
        min_distance=min_distance, sort_by=sort_by
    )
    df_mtx_sorted_mindist.to_csv(f"coevolution_{sort_by}_sorted.csv")



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import logging
logger = logging.getLogger('pypef.utils.sto2a2m')

import numpy as np
from tqdm import tqdm
from Bio import AlignIO


def convert_sto2a2m(
        sto_file: str,
        inter_gap: float,
        intra_gap: float
):
    """
    Converts alignment in format STO to A2M format.
    Removes specific sequences with inter and/or intra gaps
    over specific thresholds. More alignment type transferring
    options can be performed using the reformat.pl script of the
    HH-suite; (c) Johannes Soeding, 2012:
    https://github.com/soedinglab/hh-suite/blob/master/scripts/reformat.pl
    """
    # Generate the a2m output filename
    a2m_file = f"{sto_file.split('.sto')[0]}.a2m"

    # Load the stockholm alignment
    logger.info('Loading MSA in stockholm format...')
    sto_alignment = AlignIO.read(sto_file, 'stockholm')
    logger.info('Trimming MSA...')
    # Save this 'raw' multiple sequence alignment as numpy array 
    raw_msa = []
    for record in tqdm(sto_alignment):
        raw_msa.append(np.array(record.seq))
    raw_msa = np.array(raw_msa)

    # 1st processing step
    # Delete all positions, where WT has a gap to obtain the 'trimmed' MSA
    ungap_pos = np.where(raw_msa[0] == "-")
    msa_trimmed = np.array([np.delete(seq, ungap_pos) for seq in raw_msa])

    # 2nd processing step
    # Remove ("lower") all positions with more than 'inter_gap'*100 % gaps (columnar trimming)
    count_gaps = np.count_nonzero(msa_trimmed == '-', axis=0) / msa_trimmed.shape[0]
    lower = [idx for idx, count in enumerate(count_gaps) if count > inter_gap]
    msa_trimmed_T = msa_trimmed.T
    for idx in lower:
        msa_trimmed_T[idx] = np.char.lower(msa_trimmed_T[idx])
        # replace all columns that are "removed" due to high gap content and have an "-" element by "." 
        msa_trimmed_T[idx] = np.where(msa_trimmed_T[idx] == '-', '.', msa_trimmed_T[idx])
    msa_trimmed_inter_gap = msa_trimmed_T.T

    # 3rd processing step
    # Remove all sequences with more than 'intra_gap'*100 % gaps (line trimming)
    target_len = len(msa_trimmed_inter_gap[0])
    gap_content = (np.count_nonzero(msa_trimmed_inter_gap == "-", axis=1) + np.count_nonzero(
        msa_trimmed_inter_gap == ".", axis=1)) / target_len
    delete = np.where(gap_content > intra_gap)[0]
    msa_final = np.delete(msa_trimmed_inter_gap, delete, axis=0)
    seqs_cls = [seq_cls for idx, seq_cls in enumerate(sto_alignment) if not idx in delete]
    chunk_size = 60
    with open(a2m_file, 'w') as f:
        for i, (seq, seq_cls) in enumerate(zip(msa_final, seqs_cls)):
            if i == 0:
                f.write(f'>TARGET_SEQ\n')
            else:
                f.write('>' + seq_cls.id + '\n')
            for chunk in [seq[x:x + chunk_size] for x in range(0, len(seq), chunk_size)]:
                f.write("".join(chunk) + '\n')

    # Get number of sequences and effective sites in the alignment
    n_seqs = msa_final.shape[0]
    n_sites = sum(1 for char in msa_final[0] if char.isupper())
    logger.info(f'Generated trimmed MSA {a2m_file} in A2M format:\n'
                f'No. of sequences: {n_seqs}\n'
                f'No. of effective sites: {n_sites} (out of {target_len} sites)\n'
                f'-le --lambdae: {0.2 * (n_sites - 1):.1f}')

    return n_seqs, n_sites, target_len



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# Created on 05 October 2020
# @authors: Niklas Siedhoff, Alexander-Maurice Illig
# @contact: <n.siedhoff@biotec.rwth-aachen.de>
# PyPEF - Pythonic Protein Engineering Framework
# Released under Creative Commons Attribution-NonCommercial 4.0 International Public License (CC BY-NC 4.0)
# For more information about the license see https://creativecommons.org/licenses/by-nc/4.0/legalcode

# PyPEF – An Integrated Framework for Data-Driven Protein Engineering
# Journal of Chemical Information and Modeling, 2021, 61, 3463-3476
# https://doi.org/10.1021/acs.jcim.1c00099
# Niklas E. Siedhoff1,§, Alexander-Maurice Illig1,§, Ulrich Schwaneberg1,2, Mehdi D. Davari1,*
# 1Institute of Biotechnology, RWTH Aachen University, Worringer Weg 3, 52074 Aachen, Germany
# 2DWI-Leibniz Institute for Interactive Materials, Forckenbeckstraße 50, 52074 Aachen, Germany
# *Corresponding author
# §Equal contribution

import os

import logging
logger = logging.getLogger('pypef.utils.utils_run')

import numpy as np
import re

from pypef.utils.variant_data import (
    amino_acids, generate_dataframe_and_save_csv,
    get_basename, read_csv_and_shift_pos_ints,
    get_seqs_from_var_name, get_wt_sequence
)

from pypef.utils.learning_test_sets import (
    csv_input, drop_rows, get_variants, make_sub_ls_ts,
    make_sub_ls_ts_randomly, make_fasta_ls_ts
)
from pypef.utils.prediction_sets import (
    make_fasta_ps, make_recombinations_double, make_recombinations_triple,
    make_recombinations_quadruple, make_recombinations_quintuple,
    create_split_files, make_combinations_double_all_diverse,
    make_combinations_triple_all_diverse, make_combinations_quadruple_all_diverse,
    make_ssm_singles
)   # not yet implemented: make_combinations_double_all_diverse_and_all_positions

from pypef.utils.directed_evolution import DirectedEvolution
from pypef.utils.sto2a2m import convert_sto2a2m

from pypef.ml.regression import OneHotEncoding, AAIndexEncoding, full_aaidx_txt_path
from pypef.dca.hybrid_model import plmc_or_gremlin_encoding


def run_pypef_utils(arguments):
    if arguments['mklsts']:
        wt_sequence = get_wt_sequence(arguments['--wt'])
        t_drop = float(arguments['--drop'])

        logger.info(f'Length of provided sequence: {len(wt_sequence)} amino acids.')
        df = drop_rows(arguments['--input'], amino_acids, t_drop, arguments['--sep'], arguments['--mutation_sep'])
        no_rnd = arguments['--numrnd']

        single_variants, single_values, higher_variants, higher_values = get_variants(
            df, amino_acids, wt_sequence, arguments['--mutation_sep']
        )
        logger.info(f'Number of single variants: {len(single_variants)}.')
        if len(single_variants) == 0:
            logger.info('Found NO single substitution variants for possible recombination!')
        sub_ls, val_ls, sub_ts, val_ts = make_sub_ls_ts(
            single_variants, single_values, higher_variants, higher_values)
        logger.info('Tip: You can edit your LS and TS datasets just by '
                    'cutting/pasting between the LS and TS fasta datasets.')

        logger.info('Creating LS dataset...')
        make_fasta_ls_ts('LS.fasl', wt_sequence, sub_ls, val_ls)
        logger.info('Creating TS dataset...')
        make_fasta_ls_ts('TS.fasl', wt_sequence, sub_ts, val_ts)

        try:
            no_rnd = int(no_rnd)
        except ValueError:
            no_rnd = 0
        if no_rnd != 0:
            random_set_counter = 1
            no_rnd = int(no_rnd)
            while random_set_counter <= no_rnd:
                sub_ls, val_ls, sub_ts, val_ts = make_sub_ls_ts_randomly(
                    single_variants, single_values,
                    higher_variants, higher_values
                )
                make_fasta_ls_ts('LS_random_' + str(random_set_counter) + '.fasl', wt_sequence, sub_ls, val_ls)
                make_fasta_ls_ts('TS_random_' + str(random_set_counter) + '.fasl', wt_sequence, sub_ts, val_ts)
                random_set_counter += 1

    elif arguments['mkps']:
        wt_sequence = get_wt_sequence(arguments['--wt'])
        if not arguments['--ssm']:
            try:
                csv_file = csv_input(arguments['--input'])
            except FileNotFoundError:
                raise SystemError("If creating prediction sets ('mkps') a CSV input is "
                                  "required (if not running 'mkps --ssm').")
            t_drop = float(arguments['--drop'])
            df = drop_rows(csv_file, amino_acids, t_drop)
            drop_wt = []
            for i in range(len(df)):
                if df.iloc[i, 0] == 'WT':
                    logger.info('Dropping wild-type (WT) from DataFrame as it cannot be used for (re-)combination.')
                    drop_wt.append(i)
            df = df.drop(drop_wt).reset_index(drop=True)

            logger.info(f'Length of provided sequence: {len(wt_sequence)} amino acids.')
            single_variants, _, higher_variants, _ = get_variants(df, amino_acids, wt_sequence)
            logger.info(f'Using single substitution variants for (re-)combination. '
                        f'Number of single variants: {len(single_variants)}.')
            if len(single_variants) == 0:
                logger.info('Found NO single substitution variants for possible recombination! '
                            'No prediction files can be created!')

        if arguments['--drecomb']:
            logger.info('Creating Recomb_Double_Split...')
            for no, files in enumerate(make_recombinations_double(single_variants)):
                double_mutants = np.array(files)
                create_split_files(double_mutants, single_variants, wt_sequence, 'Recomb_Double', no)

        if arguments['--trecomb']:
            logger.info('Creating Recomb_Triple_Split...')
            for no, files in enumerate(make_recombinations_triple(single_variants)):
                triple_mutants = np.array(files)
                create_split_files(triple_mutants, single_variants, wt_sequence, 'Recomb_Triple', no)

        if arguments['--qarecomb']:
            logger.info('Beware that this step might require much disk space as PyPEF is '
                        'creating prediction files in TXT format. Creating Recomb_Quadruple_Split...')
            for no, files in enumerate(make_recombinations_quadruple(single_variants)):
                quadruple_mutants = np.array(files)
                create_split_files(quadruple_mutants, single_variants, wt_sequence, 'Recomb_Quadruple', no)

        if arguments['--qirecomb']:
            logger.info('Beware that this step might require much disk space as PyPEF is '
                        'creating prediction files in plain text format. Creating Recomb_Quintuple_Split...')
            for no, files in enumerate(make_recombinations_quintuple(single_variants)):
                quintuple_mutants = np.array(files)
                create_split_files(quintuple_mutants, single_variants, wt_sequence, 'Recomb_Quintuple', no)

        if arguments['--ddiverse']:
            logger.info('Creating Diverse_Double_Split...')
            # if functions required, uncomment the next two lines and comment the other ones
            # for no, files in enumerate(
            #     make_recombinations_double_all_diverse_and_all_positions(wt_sequence, amino_acids)):
            for no, files in enumerate(make_combinations_double_all_diverse(single_variants, amino_acids)):
                doubles = np.array(files)
                create_split_files(doubles, single_variants, wt_sequence, 'Diverse_Double', no + 1)

        if arguments['--tdiverse']:
            logger.info('Beware that this step might require much disk space as PyPEF is '
                        'creating prediction files in plain text format. Creating Diverse_Triple_Split... ')
            for no, files in enumerate(make_combinations_triple_all_diverse(single_variants, amino_acids)):
                triples = np.array(files)
                create_split_files(triples, single_variants, wt_sequence, 'Diverse_Triple', no + 1)

        if arguments['--qdiverse']:
            logger.info('Beware that this step might require much disk space as PyPEF is '
                       'creating prediction files in plain text format. Creating Diverse_Quadruple_Split...')
            for no, files in enumerate(make_combinations_quadruple_all_diverse(single_variants, amino_acids)):
                quadruples = np.array(files)
                create_split_files(quadruples, single_variants, wt_sequence, 'Diverse_Quadruple', no + 1)

        if arguments['--ssm']:
            singles = make_ssm_singles(wt_sequence, amino_acids)
            make_fasta_ps('ssm_singles.fasta', wt_sequence, np.array(singles))

        if True not in [
            arguments['--drecomb'], arguments['--trecomb'], arguments['--qarecomb'],
            arguments['--qirecomb'], arguments['--ddiverse'], arguments['--tdiverse'],
            arguments['--qdiverse'], arguments['--ssm']
        ]:
            logger.info(f'\nMaking prediction set fasta file from {csv_file}...\n')
            make_fasta_ps(
                filename=f'{get_basename(csv_file)}_prediction_set.fasta',
                wt=wt_sequence,
                substitutions=tuple(list(single_variants) + list(higher_variants))
            )

    # Metropolis-Hastings-driven directed evolution, similar to Biswas et al.:
    # Low-N protein engineering with data-efficient deep learning,
    # see https://github.com/ivanjayapurna/low-n-protein-engineering/tree/master/directed-evo
    elif arguments['directevo']:
        if arguments['hybrid'] or arguments['--encoding'] == 'dca':
            dca_encoder = arguments['--params']
            if arguments['ml']:
                ml_or_hybrid = 'ml'
            else:
                ml_or_hybrid = 'hybrid'
        else:
            dca_encoder = None
            ml_or_hybrid = 'ml'
        # Prediction using a saved model Pickle file specific AAindex used for encoding
        # Model saved in Pickle file also for DCA-based encoding, a default file name
        logger.info('Not counting WT as variant in directed evolution '
                    'as it cannot be used for (re-)combination.')
        path = os.getcwd()
        try:
            # "temperature" parameter: determines sensitivity of Metropolis-Hastings acceptance criteria
            temp = float(arguments['--temp'])
            # how many subsequent mutation trials per simulated evolution trajectory
            num_iterations = int(arguments['--numiter'])
            # how many separate evolution trajectories to run
            num_trajectories = int(arguments['--numtraj'])
        except ValueError:
            raise ValueError("Define flags 'numiter' and 'numtraj' as integer and 'temp' as float.")
        s_wt = get_wt_sequence(arguments['--wt'])
        y_wt = arguments['--y_wt']
        negative = arguments['--negative']
        # Metropolis-Hastings-driven directed evolution on single mutant position csv data
        usecsv = arguments['--usecsv']
        if usecsv:
            csv_file = csv_input(arguments['--input'])
            t_drop = float(arguments['--drop'])
            logger.info(f'Length of provided sequence: {len(s_wt)} amino acids.')
            df = drop_rows(csv_file, amino_acids, t_drop)
            drop_wt = []
            for i in range(len(df)):
                if df.iloc[i, 0] == 'WT':
                    logger.info('Using fitness value (y_WT) for wild-type (WT) as specified in CSV.')
                    drop_wt.append(i)
                    y_wt = df.iloc[i, 1]
            df = df.drop(drop_wt).reset_index(drop=True)
            single_variants, single_values, higher_variants, higher_values = \
                get_variants(df, amino_acids, s_wt)
            logger.info(f'Number of single variants: {len(single_variants)}.')
            if len(single_variants) == 0:
                logger.info('Found NO single substitution variants for possible recombination!')
            single_vars, single_ys = list(single_variants), list(single_values)  # only tuples to lists

        else:
            single_vars = None  # What happens now? (Full diverse?)
        # Metropolis-Hastings-driven directed evolution on single mutant .csv amino acid substitution data
        csvaa = arguments['--csvaa']  # only use identified substitutions --> taken from CSV file
        logger.info('Running evolution trajectories and plotting...')
        DirectedEvolution(
            ml_or_hybrid=ml_or_hybrid,
            encoding=arguments['--encoding'],
            s_wt=s_wt,
            y_wt=y_wt,
            single_vars=single_vars,
            num_iterations=num_iterations,
            num_trajectories=num_trajectories,
            amino_acids=amino_acids,
            temp=temp,
            path=path,
            model=arguments['--model'],
            no_fft=arguments['--nofft'],
            dca_encoder=dca_encoder,
            usecsv=usecsv,
            csvaa=csvaa,
            negative=negative
        ).plot_trajectories()


    elif arguments['sto2a2m']:
        convert_sto2a2m(
            sto_file=arguments['--sto'],
            inter_gap=arguments['--inter_gap'],
            intra_gap=arguments['--intra_gap']
        )

    elif arguments['reformat_csv']:
        read_csv_and_shift_pos_ints(
            infile=arguments['--input'],
            offset=0,
            col_sep=arguments['--sep'],
            substitution_sep=arguments['--mutation_sep']
        )

    elif arguments['shift_pos']:
        read_csv_and_shift_pos_ints(
            infile=arguments['--input'],
            offset=arguments['--offset'],
            col_sep=arguments['--sep'],
            substitution_sep=arguments['--mutation_sep']
        )

    elif arguments['encode']:  # sole parallelized task for utils for DCA encoding
        df = drop_rows(arguments['--input'], amino_acids, arguments['--drop'])
        wt_sequence = get_wt_sequence(arguments['--wt'])
        logger.info(f'Length of provided sequence: {len(wt_sequence)} amino acids.')
        single_variants, single_values, higher_variants, higher_values = get_variants(
            df, amino_acids, wt_sequence)
        variants = list(single_variants) + list(higher_variants)
        ys_true = list(single_values) + list(higher_values)
        variants, ys_true, sequences = get_seqs_from_var_name(wt_sequence, variants, ys_true)
        assert len(variants) == len(ys_true) == len(sequences)
        logger.info('Encoding variant sequences...')

        if arguments['--encoding'] == 'dca':
            threads = abs(arguments['--threads']) if arguments['--threads'] is not None else 1
            threads = threads + 1 if threads == 0 else threads
            logger.info(f'Using {threads} thread(s) for running...')
            xs, variants, sequences, ys_true, x_wt, model, model_type = plmc_or_gremlin_encoding(
                variants=variants,
                sequences=sequences,
                ys_true=ys_true,
                params_file=arguments['--params'],
                substitution_sep=arguments['--mutation_sep'],
                threads=threads,
                verbose=True
            )
            assert len(xs) == len(variants) == len(ys_true)

            if variants[0][0] != variants[0][-1]:  # WT is required for DCA-based hybrid modeling
                if arguments['--y_wt'] is not None:
                    y_wt = arguments['--y_wt']
                else:
                    y_wt = 1
                # better using re then: wt = variants[0][0] + str(variants[0][1:-1] + variants[0][0])
                wt = variants[0][0] + re.findall(r"\d+", variants[0])[0] + variants[0][0]
                variants = list(variants)
                variants.insert(0, wt)  # inserting WT at pos. 0
                xs = list(xs)
                xs.insert(0, list(x_wt.flatten()))
                ys_true = list(ys_true)
                ys_true.insert(0, y_wt)  # set WT fitness to 1 or use arguments y_wt?

        elif arguments['--encoding'] == 'onehot':
            onehot_encoder = OneHotEncoding(sequences)
            xs = onehot_encoder.collect_encoded_sequences()

        elif arguments['--encoding'] == 'aaidx':
            if arguments['--model'] is None:
                raise SystemError(
                    "Define the AAindex to use for encoding with the "
                    "flag --model AAINDEX, e.g.: --model CORJ870104."
                )
            aa_index_encoder = AAIndexEncoding(
                full_aaidx_txt_path(arguments['--model'] + '.txt'), sequences
            )
            x_fft, x_raw = aa_index_encoder.collect_encoded_sequences()
            if arguments['--nofft']:
                xs = x_raw
            else:
                xs = x_fft

        else:
            raise SystemError("Unknown encoding option.")
        logger.info(f'{len(variants)} variants (plus inserted WT) remained after encoding. '
                    f'Saving to encoding CSV file...')
        generate_dataframe_and_save_csv(  # put WT at pos. 0 for hybrid low_N or extrapolation
            variants=variants,
            sequence_encodings=xs,
            fitnesses=ys_true,
            csv_file=arguments['--input'],
            encoding_type=arguments['--encoding']
        )
